#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   getNodeForceAtArea.py
#
#       nodeForceが等分布荷重となるように設定する。
#       nodeGroupは、face、edge、pointsに対して適用できる。
#       （nodeForceの合計＝1.0で取得する）
#       要素の種類は、以下の種類に対応。
#           四面体（tetra）  1次2次
#           六面体（hexa）   1次2次
#           五面体（prism）  1次2次
#           シェル三角形     1次（731、761）
#           シェル四角形     1次（741、781）
#
#   15/12/28    新規作成
#   16/01/04    2次要素の場合、中間節点のみに設定
#      03/27    シェル要素1次（731、741）を追加
#      04/08    シェル要素1次（761、781）を追加
#      12/17    シェル要素の面積、ビーム要素の長さの取得を追加
#      12/21    getElementWithNode:高速化の為修正
#      12/25    setAllNodesList:バグ修正（nodeListの取りこぼしがあった。）
#


import pyFistr

nodeList = []
elementList = []
areaSize = []
nodeElmList = []

#
#  setAllNodesList
#    nodeのdataを取得
def setAllNodesList(lines):
    global nodeList

    def getNodeData(lines, i):
        #!NODEまでskip
        si = len(lines)
        while i < len(lines):
            if lines[i][:len("!NODE")] == "!NODE":
                si = i
                break
            i += 1
        if si == len(lines):
            return [[], si]

        #nodeのdataを取得
        nodes = []
        i = si + 1
        while i < len(lines):
            if lines[i][0] != "!":
                words = lines[i].split(",")
                data = [int(words[0]), 
                        float(words[1]), float(words[2]), float(words[3])]
                nodes.append(data)
            else:
                break
            i += 1
        return [nodes, i]

    #nodeのdataを取得
    nodesData = []
    i = 0
    while i < len(lines):
        [nodes, i] = getNodeData(lines, i)
        if len(nodes) != 0:
            nodesData += nodes
        #i += 1
    #nodeDataをlistに保存
    nodesData.sort()
    maxNo = nodesData[-1][0]
    nodeList = []
    for i in range(maxNo+1):
        nodeList.append([])
    for data in nodesData:
        node = data[0]       
        loc = data[1:]
        nodeList[node] = loc

#
#  setAllElementList
#    全てのelementのdata取得
def setAllElementList(lines):
    global elementList, nodeElmList

    def getElementData(lines, i):
        #!ELEMENTまでskip
        si = len(lines)
        while i < len(lines):
            if lines[i][:len("!ELEMENT")] == "!ELEMENT":
                elType = pyFistr.getValName(lines[i], "TYPE")
                si = i
                break
            i += 1
        if si == len(lines):
            return [[], si]

        #elementのdataを取得
        elements = []
        i = si + 1
        while i < len(lines):
            if lines[i][0] != "!":
                words = lines[i].split(",")
                data = []
                for word in words:
                    try:
                        a = int(word)
                        data.append(a)
                    except:
                        pass
                elements.append(data)
            else:
                break
            i += 1
        return [[elType, elements], i]

    #elementのdataを取得
    elementsData = []
    i = 0
    while i < len(lines):
        [elementsType, i] = getElementData(lines, i)
        if len(elementsType) != 0:
            elementsData.append(elementsType)
        i += 1
    #nodeDataをlistに保存
    elNums = []
    for elements in elementsData:
        for elNo in elements[1]:
            elNums.append(elNo)
    elNums.sort()
    maxNo = elNums[-1][0]
    elementList = []
    for i in range(maxNo+1):
        elementList.append([])
    for elements in elementsData:
        elType = elements[0]
        for data in elements[1]:
            elNo = data[0]
            nodes = data[1:]
            elementList[elNo] = [elType, nodes]

#
#  setAllNodeElmList
#    nodeNoをindexとしてelementのデータを保存
def setAllNodeElmList(lines):
    global nodeElmList

    def getElementData(lines, i):
        #!ELEMENTまでskip
        si = len(lines)
        while i < len(lines):
            if lines[i][:len("!ELEMENT")] == "!ELEMENT":
                elType = pyFistr.getValName(lines[i], "TYPE")
                si = i
                break
            i += 1
        if si == len(lines):
            return [[], si]

        #elementのdataを取得
        elements = []
        i = si + 1
        while i < len(lines):
            if lines[i][0] != "!":
                words = lines[i].split(",")
                data = []
                for word in words:
                    try:
                        a = int(word)
                        data.append(a)
                    except:
                        pass
                elements.append(data)
            else:
                break
            i += 1
        return [[elType, elements], i]


    #配列を準備
    nodeElmList = []
    for i in range(len(nodeList) + 1):
        nodeElmList.append([])
    #要素の内容を取得
    i = 0
    while i < len(lines):
        [elementsType, i] = getElementData(lines, i)
        if len(elementsType) != 0:
            [elType, elements] = elementsType
            for element in elements:
                elNo = element[0]
                nodes = element[1:]
                n = 0
                for node in nodes:
                    nodeElmList[node] += [[elType, elNo, n]]
                    n += 1
        i += 1

#
#  getNodeNo
#    nodeNoを取得
def getNodeNo(lines, ngrpName):
    nodeNo = []
    flag = 0
    for i in range(len(lines)):
        line = lines[i]
        if line[0] == "!":
            flag = 0
        if line[:len("!NGROUP")] == "!NGROUP":
            words = pyFistr.deleteSp(line).split(",")
            for word in words:
                if word[:len("NGRP=")] == "NGRP=":
                    name = word.split("=")[1]
                    if name == ngrpName:
                        flag = 1
        else:
            if flag == 1:
                #nodeNoを取得
                words = line.split(",")
                for word in words:
                    try:
                        nodeNo.append(int(word))
                    except:
                        pass
    return nodeNo

#
#  getElementWithNode
#    nodeを含むelementを取得
#    [[<elType>, <elNo>], [nodeOrder]]を取得
#    [["761", 2942],      [0, 1, 2]]
def getElementWithNode(nodeNo):
    #配列を準備
    elementNodes = []
    for i in range(len(elementList)):
        elementNodes.append([])
    #elementを取得
    """
    for node in nodeNo:
        for i in range(len(elementList)):
            if len(elementList[i]) > 0:
                elNodes = elementList[i][1]
                if len(elNodes) > 0:
                    for n in range(len(elNodes)):
                        if node == elNodes[n]:
                            elementNodes[i] += [n]
                            break
    """
    for node in nodeNo:
        elms = nodeElmList[node]
        for elm in elms:
            [elType, elNo, n] = elm
            elementNodes[elNo] += [n]
    #データ整形
    elNodes = []
    for i in range(len(elementNodes)):
        nodes = elementNodes[i]
        if len(nodes) > 0:
            nodes.sort()
            data = [[elementList[i][0], i], nodes]
            elNodes.append(data)
    return elNodes

#
#  lengthPP
#    線分の長さを取得
def lengthPP(loc1, loc2):
    locx = loc1[0] - loc2[0]
    locy = loc1[1] - loc2[1]
    locz = loc1[2] - loc2[2]
    l = pow((locx*locx + locy*locy + locz*locz), 0.5)
    return l    

#
#  calcTriangleArea
#    三角形の面積を計算
def calcTriangleArea(loc1, loc2, loc3):
    #各辺の長さを計算
    a = lengthPP(loc1, loc2)
    b = lengthPP(loc2, loc3)
    c = lengthPP(loc3, loc1)
    s = (a + b + c)/2.0
    area = pow((s*(s - a)*(s - b)*(s - c)), 0.5)
    return area

#
#  calcArea341
#    三角形の面積を取得
def calcArea341(elNo, nodes):
    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    nodeNo3 = elNodes[nodes[2]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    loc3 = nodeList[nodeNo3]
    area = calcTriangleArea(loc1, loc2, loc3)
    return [area, [nodeNo1, nodeNo2, nodeNo3]]

#
#  calcArea342
#    三角形の面積を取得
def calcArea342(elNo, nodes):
    #エラーチェック
    if len(nodes) < 6:
        area = 0.0
        return [area, []]

    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    nodeNo3 = elNodes[nodes[2]]
    nodeNo4 = elNodes[nodes[3]]
    nodeNo5 = elNodes[nodes[4]]
    nodeNo6 = elNodes[nodes[5]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    loc3 = nodeList[nodeNo3]
    #面積計算
    area = calcTriangleArea(loc1, loc2, loc3)
    return [area, [nodeNo1, nodeNo2, nodeNo3, nodeNo4, nodeNo5, nodeNo6]]

#
#  calcArea361
#    faceの面積を取得
def calcArea361(elNo, nodes):
    if len(nodes) < 4:
        area = 0.0
        return [area, []]
    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    nodeNo3 = elNodes[nodes[2]]
    nodeNo4 = elNodes[nodes[3]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    loc3 = nodeList[nodeNo3]
    loc4 = nodeList[nodeNo4]
    #面積計算
    area  = calcTriangleArea(loc1, loc2, loc3)
    area += calcTriangleArea(loc3, loc4, loc1)        
    return [area, [nodeNo1, nodeNo2, nodeNo3, nodeNo4]]

#
#  calcArea362
#    faceの面積を取得
def calcArea362(elNo, nodes):
    if len(nodes) < 8:
        area = 0.0
        return [area, []]
    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    nodeNo3 = elNodes[nodes[2]]
    nodeNo4 = elNodes[nodes[3]]
    nodeNo5 = elNodes[nodes[4]]
    nodeNo6 = elNodes[nodes[5]]
    nodeNo7 = elNodes[nodes[6]]
    nodeNo8 = elNodes[nodes[7]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    loc3 = nodeList[nodeNo3]
    loc4 = nodeList[nodeNo4]
    #面積計算
    area  = calcTriangleArea(loc1, loc2, loc3)
    area += calcTriangleArea(loc3, loc4, loc1)        
    return [area, [nodeNo1, nodeNo2, nodeNo3, nodeNo4, nodeNo5, nodeNo6, nodeNo7, nodeNo8]]

#
#  calcArea351
#    faceの面積を計算
def calcArea351(elNo, nodes):
    if len(nodes) == 3:
        sNodes = calcArea341(elNo, nodes)
    elif len(nodes) == 4:
        sNodes = calcArea361(elNo, nodes)
    else:
        area = 0.0
        sNodes = [area, []]
    return sNodes

#
#  calcArea352
#    faceの面積を計算
def calcArea352(elNo, nodes):
    if len(nodes) == 6:
        sNodes = calcArea342(elNo, nodes)
    elif len(nodes) == 8:
        sNodes = calcArea362(elNo, nodes)
    else:
        area = 0.0
        sNodes = [area, []]
    return sNodes

#
#  calcArea731
#    shellFaceの面積を計算
def calcArea731(elNo, nodes):
    if len(nodes) == 3:
        sNodes = calcArea341(elNo, nodes)
    elif len(nodes) == 6:
        sNodes = calcArea341(elNo, nodes[:3])
    return sNodes

#
#  calcArea741
#    shellFaceの面積を計算
def calcArea741(elNo, nodes):
    if len(nodes) == 4:
        sNodes = calcArea361(elNo, nodes)
    elif len(nodes) == 8:
        sNodes = calcArea361(elNo, nodes[:4])
    return sNodes

#
#  calcArea732
#    shell2Faceの面積を計算
def calcArea732(elNo, nodes):
    sNodes = calcArea342(elNo, nodes)

#
#  calcArea742
#    shell2Faceの面積を計算
def calcArea742(elNo, nodes):
    sNodes = calcArea362(elNo, nodes)

#
#  getAreaWithNodes
#    nodeを含む領域の面積を取得
def getAreaWithNodes(elNodes):
    areaNodes = []
    for elNode in elNodes:
        if len(elNode[1]) >= 3:
            elType = elNode[0][0]       #要素タイプ
            elNo = elNode[0][1]         #要素No
            node = elNode[1]            #nodeNoの組
            if elType == "341":
                sNodes = calcArea341(elNo, node)
            elif elType == "342":
                sNodes = calcArea342(elNo, node)
            elif elType == "361":
                sNodes = calcArea361(elNo, node)
            elif elType == "362":
                sNodes = calcArea362(elNo, node)
            elif elType == "351":
                sNodes = calcArea351(elNo, node)
            elif elType == "352":
                sNodes = calcArea352(elNo, node)
            elif elType == "732":
                sNodes = calcArea732(elNo, node)
            elif elType == "742":
                sNodes = calcArea742(elNo, node)
            elif elType == "731" or elType == "761":
                sNodes = calcArea731(elNo, node)
            elif elType == "741" or elType == "781":
                sNodes = calcArea741(elNo, node)
            area = sNodes[0]
            if area != 0.0:
                areaNodes.append(sNodes)
    return areaNodes

#
#  calcNodeForce
#    節点荷重をセット（荷重合計＝1.0）
def calcNodeForce(areaNodes, nodeNo):
    #全面積を取得
    area = 0.0
    for [s, nodes] in areaNodes:
        area += s
    #配列を準備
    setForce = []
    for node in nodeNo:
        setForce.append(0.0)

    #1次要素の等分布荷重を算出
    if len(nodes) == 3:
        #三角形1次要素
        for [s, nodes] in areaNodes:
            force = s / area / len(nodes)
            for setNode in nodes:
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += force
                        break
    elif len(nodes) == 4:
        #四角形1次要素
        for [s, nodes] in areaNodes:
            force = s / area / len(nodes)
            #nodesに荷重を配分
            for setNode in nodes:
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += force
                        break 
    #2次要素の場合
    elif len(nodes) == 6:
        #三角形2次要素（コーナ:0、中間:1/3）
        for [s, nodes] in areaNodes:
            force = s / area / 3
            #中間節点のみへ配分
            for n in [3,4,5]:
                setNode = nodes[n]
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += force
                        break
    elif len(nodes) == 8:
        #四角形2次要素（コーナ:-1/12、中間:1/3）
        for [s, nodes] in areaNodes:
            force = s / area / 12.0
            #コーナ節点への配分
            for n in [0,1,2,3]:
                setNode = nodes[n]
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += -1.0 * force
                        break
            #中間節点への配分
            for n in [4,5,6,7]:
                setNode = nodes[n]
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += 4.0 * force
                        break
    #面積算出
    a = 0.0
    for [s, nodes] in areaNodes:
        a += s
    #荷重係数算出（合計＝1.0）
    tForce = 0.0
    for force in setForce:
        tForce += force
    print _(u"面に等分布荷重設定　面の面積:") + str(a) + "  荷重係数:" + str(tForce)
    return setForce

#
#  calcLine341
#    lineの長さを取得
def calcLine341(elNo, nodes):
    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    #線分の長さを取得
    lx = (loc1[0]-loc2[0])
    ly = (loc1[1]-loc2[1])
    lz = (loc1[2]-loc2[2])
    line = pow((lx*lx + ly*ly + lz*lz), 0.5)
    return [line, [nodeNo1, nodeNo2]]

#
#  calcLine342
#    lineの長さを取得
def calcLine342(elNo, nodes):
    if len(nodes) != 3:
        line = 0.0
        return [line, []]
    element = elementList[elNo]
    elNodes = element[1]
    #node番号を取得
    nodeNo1 = elNodes[nodes[0]]
    nodeNo2 = elNodes[nodes[1]]
    nodeNo3 = elNodes[nodes[2]]
    #座標を取得
    loc1 = nodeList[nodeNo1]
    loc2 = nodeList[nodeNo2]
    #線分の長さを取得
    lx = (loc1[0]-loc2[0])
    ly = (loc1[1]-loc2[1])
    lz = (loc1[2]-loc2[2])
    line = pow((lx*lx + ly*ly + lz*lz), 0.5)
    return [line, [nodeNo1, nodeNo2, nodeNo3]]    

#
#  calcLine361
#    線分の長さを取得
def calcLine361(elNo, nodes):
    lNodes = calcLine341(elNo, nodes)
    return lNodes

#
#  calcLine362
#    線分の長さを取得
def calcLine362(elNo, nodes):
    lNodes = calcLine342(elNo, nodes)
    return lNodes

#
#  calcLine351
#    線分の長さを取得
def calcLine351(elNo, nodes):
    lNodes = calcLine341(elNo, nodes)
    return lNodes

#
#  calcLine352
#    線分の長さを取得
def calcLine352(elNo, nodes):
    lNodes = calcLine342(elNo, nodes)
    return lNodes

#
#  calcLine731
#    線分の長さを取得
def calcLine731(elNo, nodes):
    lNodes = calcLine341(elNo, nodes)
    return lNodes

#
#  calcLine741
#    線分の長さを取得
def calcLine741(elNo, nodes):
    lNodes = calcLine341(elNo, nodes)
    return lNodes

#
#  calcLine611
#    beamの線分の長さを取得
def calcLine611(elNo, nodes):
    lNodes = calcLine341(elNo, nodes)
    return lNodes

#
#  getLineWithNodes
#    nodeを含む線分の長さを取得
def getLineWithNodes(elNodes):
    lineNodes = []
    for elNode in elNodes:
        if len(elNode[1]) >= 2:
            elType = elNode[0][0]       #要素タイプ
            elNo = elNode[0][1]         #要素No
            node = elNode[1]            #nodeNoの組
            if elType == "341":
                lNodes = calcLine341(elNo, node)
            elif elType == "342":
                lNodes = calcLine342(elNo, node)
            elif elType == "361":
                lNodes = calcLine361(elNo, node)
            elif elType == "362":
                lNodes = calcLine362(elNo, node)
            elif elType == "351":
                lNodes = calcLine351(elNo, node)
            elif elType == "352":
                lNodes = calcLine352(elNo, node)
            elif elType == "731" or elType == "761":
                lNodes = calcLine731(elNo, node)
            elif elType == "741" or elType == "781":
                lNodes = calcLine741(elNo, node)
            elif elType == "611" or elType == "641":
                lNodes = calcLine611(elNo, node)
            line = lNodes[0]
            if line != 0.0:
                lineNodes.append(lNodes)
    if len(lineNodes) != 0:
        #取得した線分をチェックする（違う要素で同じ線分が含まれている場合がある。）
        #  check用の配列checkNodeにコピー
        checkNodes = []
        for i in range(len(lineNodes)):
            lineNode = lineNodes[i]
            checkNodes.append([lineNode, i])
            checkNodes[-1][0][1].sort()         #節点Noをsortする
        #　データ全体をsortする。
        checkNodes.sort()
        newNodes = []
        addNode = []
        for i in range(len(checkNodes)):
            checkNode = checkNodes[i][0]
            if checkNode != addNode:
                #節点をsortする前のデータを保存
                n = checkNodes[i][1]
                newNodes.append(lineNodes[n])
                addNode = checkNode
        return newNodes
    else:
        return lineNodes
    
#
#  calcNodeForceLine
#    線分の荷重を計算
def calcNodeForceLine(lineNodes, nodeNo):
    line = 0.0
    for [l, nodes] in lineNodes:
        line += l
    setForce = []
    for node in nodeNo:
        setForce.append(0.0)
    if len(nodes) == 2:
        #1次要素（均等に分配）
        for [l, nodes] in lineNodes:
            force = l / line / len(nodes)
            for setNode in nodes:
                for i in range(len(nodeNo)):
                    node = nodeNo[i]
                    if setNode == node:
                        setForce[i] += force
    elif len(nodes) == 3:
        #2次要素（コーナ節点:1、中間節点:4）
        for [l, nodes] in lineNodes:
            force = l / line / 6.0
            for n in range(len(nodes)):
                setNode = nodes[n]
                if n < 2:
                    for i in range(len(nodeNo)):
                        node = nodeNo[i]
                        if setNode == node:
                            setForce[i] += force
                else:
                    for i in range(len(nodeNo)):
                        node = nodeNo[i]
                        if setNode == node:
                            setForce[i] += 4.0 * force
    l = 0.0
    for [line, nodes] in lineNodes:
        l += line
    tForce = 0.0
    for force in setForce:
        tForce += force
    print _(u"線分に等分布荷重設定　線分の全長:") + str(l) + "  荷重係数:" + str(tForce)
    return setForce

#
#  calcNodeForcePoint
#    各pointの均等に荷重をセット
def calcNodeForcePoint(nodeNo):
    force = 1.0/float(len(nodeNo))
    setForce = []
    for i in range(len(nodeNo)):
        setForce.append(force)
    print _(u"点に等分布荷重設定　点の数:") + str(len(nodeNo))
    return setForce

#
#  getNodeForce
#  ------------
#
def getNodeForce(lines, ngrpName):
    #全てのnode、elementのdataを取得
    setAllNodesList(lines)
    setAllElementList(lines)
    setAllNodeElmList(lines)
    #nodeNoを取得
    print _(u"節点座標を取得中...")
    nodeNo = getNodeNo(lines, ngrpName)
    #nodeを含むelementを取得
    elNodes = getElementWithNode(nodeNo)
    #領域の面積と領域のnodeNoを取得
    areaNodes = getAreaWithNodes(elNodes)
    if len(areaNodes) > 0:
        #面積に応じた荷重を計算
        nodeForce = calcNodeForce(areaNodes, nodeNo)
    else:
        #線分の長さに応じた荷重を計算
        lineNodes = getLineWithNodes(elNodes)
        if len(lineNodes) > 0:
            nodeForce = calcNodeForceLine(lineNodes, nodeNo)
        else:
            #節点に均等に荷重をかける
            nodeForce = calcNodeForcePoint(nodeNo)
    return [nodeNo, nodeForce]


if __name__=="__main__":
    fileName = "FistrModel.msh"
    lines = []
    f = open(fileName)
    for line in f.readlines():
        if line != "\n":
            lines.append(line)
    f.close()
    ngrpName = "point"
    [nodeNo, nodeForce] = getNodeForce(lines, ngrpName)
    print nodeNo
    print nodeForce

